from enum import Enum

class JobStatus(Enum):
    IN_PROGRESS = 2
    COMPLETED = 3
    FAILED = 4
    ERROR = 5

class ComputeStep(Enum):
    TOKENIZE = 0
    EMBED = 1
    LAYER = 2
    NORM = 3
    HEAD = 4

class ModelPartType(Enum):
    EMBED = 0
    NORM = 1
    HEAD = 2
    LAYER = 3



import logging
from threading import Thread
from distributed_state_network import DSNodeServer

from language_pipes.job_manager import JobManager
from language_pipes.job_receiver import JobReceiver
from language_pipes.handlers.oai import OAIHttpServer

from language_pipes.util import stop_thread
from language_pipes.config import LpConfig

def serve(httpd):
    httpd.serve_forever()

class LanguagePipes:
    router: DSNodeServer
    
    job_manager: JobManager
    job_receiver: JobReceiver

    oai_server: OAIHttpServer
    oai_thread: Thread
    
    config: LpConfig

    def __init__(
        self, 
        version: str,
        config: LpConfig
    ):
        self.config = config
        self.set_logging_level(self.config.logging_level, self.config.router.node_id)
        self.job_manager = None
        self.router = DSNodeServer.start(self.config.router, self.print_pipes, self.print_pipes)
        self.job_manager = JobManager(config.app_dir, self.router.node, self.config.processor)
        self.job_receiver = JobReceiver(
            self.config.processor, 
            self.router.node, 
            self.job_manager.get_pipe, 
            self.job_manager.get_end_model, 
            self.job_manager.get_pending_job,
            self.job_manager.restart_job
        )
        if self.config.oai_port is not None:
            self.start_oai()

    def print_pipes(self):
        if self.job_manager is None:
            return
        self.job_manager.print_pipes()

    def start_oai(self):
        self.oai_server = OAIHttpServer(self.config.oai_port, self.job_manager.complete)
        self.oai_thread = Thread(target=self.oai_server.serve_forever, args=())
        self.oai_thread.start()
        self.job_manager.logger.info(f"OpenAI Server started on port {self.config.oai_port}")

    def set_logging_level(self, logging_level: str, router_id: str):
        level = getattr(logging, logging_level.upper(), None)
        if level is None:
            raise ValueError(f"Invalid logging level: {logging_level}")
        logging.basicConfig(level=level)

    def router_port(self) -> int:
        return self.config.router.port

    def stop(self):
        self.job_manager.stop()
        self.job_receiver.stop()
        self.router.stop()
        if self.config.oai_port is not None:
            self.oai_server.shutdown()
            stop_thread(self.oai_thread)



import os
import requests
from threading import Thread
from typing import Callable, List, Optional
from pathlib import Path
from uuid import uuid4

from transformers import AutoTokenizer
from transformers.models.auto import AutoConfig
from distributed_state_network import DSNode

from language_pipes.util.meta import MetaPipe
from language_pipes.util.chat import ChatMessage
from language_pipes.job_manager.enums import JobStatus
from language_pipes.llm_model import LlmModel
from language_pipes.job_manager.layer_job import LayerJob
from language_pipes.job_manager.job import Job

class Pipe:
    pipe_id: str
    model_id: str
    segments: List[LlmModel]

    router: DSNode
    tokenizer: Callable
    get_job_port: Callable[[str], Optional[int]]
    complete_job: Callable[[Job], None]
    update_job: Callable[[Job], None]
    model_num_hidden_layers: int

    def __init__(
            self, 
            router: DSNode,
            pipe_id: Optional[str],
            model_id: str,
            app_dir: str,
            get_job_port: Callable[[str], Optional[int]],
            complete_job: Callable[[Job], None],
            update_job: Callable[[Job], None],
            restart_job: Callable[[Job], None]
        ):
        self.get_job_port = get_job_port
        self.complete_job = complete_job
        self.update_job = update_job
        self.restart_job = restart_job
        self.router = router
        self.model_id = model_id
        model_dir = str(Path(app_dir) / 'models' / model_id / 'data')
        self.model_num_hidden_layers = AutoConfig.from_pretrained(model_dir).num_hidden_layers
        
        if pipe_id is None:
            self.pipe_id = str(uuid4())
        else:
            self.pipe_id = pipe_id

        self.segments = []
        self.tokenizer = lambda: AutoTokenizer.from_pretrained(os.path.join('models', model_id, 'data'))

    def raise_exception(self, msg: str):
        self.router.logger.exception(msg)
        raise Exception(msg)

    def send_job(self, job: LayerJob, router_id: str):
        try:
            ip = self.router.connection_from_node(router_id).address
            port = self.get_job_port(router_id)
            if port is None:
                self.raise_exception(f"SEND JOB => Could not find pipe {self.pipe_id} for {router_id}")

            self.router.logger.info(f'Sending job {job.job_id} to {router_id}')
            def send(url: str, data: bytes):
                try:
                    res = requests.post(url, data=data, headers={'Content-Type': 'application/octet-stream'})
                    if res.status_code != 200 or res.content == b'DOWN':
                        self.raise_exception(f"SEND JOB => bad response from {router_id}")
                except:
                    self.raise_exception(f"SEND JOB => Could not connect to {router_id}")
            Thread(target=send, args=(f'http://{ip}:{port}', job.to_bytes(), )).start()
        except:
            self.restart_job(job)

    def tokenize(self, prompt: Optional[str], messages: List[ChatMessage]) -> List[int]:
        tokenizer: AutoTokenizer = self.tokenizer()
        if prompt is None:
            prompt = tokenizer.apply_chat_template([m.to_json() for m in messages], tokenize=False, add_generation_prompt=True, chat_template=tokenizer.chat_template)
        return [int(t) for t in tokenizer.encode(prompt, return_tensors='pt')[0].numpy()]

    def get_layer(self, layer: int, need_physical: bool = False) -> Optional[LlmModel]:
        for segment in self.segments:
            if segment.start_layer == layer and (not need_physical or not segment.virtual):
                return segment
        return None
    
    def get_computed(self):
        return self.segments[0].computed

    def sort_segments(self):
        self.segments = sorted(self.segments, key=lambda x: x.start_layer)

    def is_complete(self):
        self.sort_segments()
        current_layer = 0
        for s in self.segments:
            if s.start_layer == current_layer:
                current_layer = s.end_layer + 1

        return current_layer == self.model_num_hidden_layers

    def print(self):
        self.router.logger.info(f'''
=================================
Pipe Status:
Model ID: {self.model_id}
Pipe: {self.pipe_id}
Segments: {', '.join([s.router_id for s in self.segments])}
Embed: {not self.get_embed() is not None}
Head: {not self.get_head() is not None}
End Layer: {self.segments[-1].end_layer}
Complete: {self.is_complete()}
=================================
''')

    def peers(self) -> List[str]:
        peers: List[str] = []
        for segment in self.segments:
            if segment.router_id not in peers:
                peers.append(segment.router_id)
        return peers

    def model_for_job(self, job: Job, need_physical: bool = False) -> Optional[LlmModel]:
        return self.get_layer(job.current_layer, need_physical)

    def process_job(
            self,
            job: Job
        ):
        while True:
            model_for_job = self.model_for_job(job)
            if model_for_job is None:
                if job.status == JobStatus.COMPLETED:
                    if job.router_id == self.router.config.node_id:
                        res_tokens = job.input_id_tensor()
                        job.result = self.tokenizer().decode(res_tokens[job.prompt_tokens:])
                        self.complete_job(job)
                    else:
                        self.send_job(job, job.router_id)
                else:
                    self.restart_job(job)
                return
            
            if model_for_job.virtual:
                self.send_job(job, model_for_job.router_id)
                return
            
            model_for_job.process_job(job)

    @staticmethod
    def from_meta(
        meta_pipe: MetaPipe, 
        hosted_models: List[LlmModel], 
        router: DSNode,
        app_dir: str,
        get_job_port: Callable[[str], Optional[int]],
        complete_job: Callable[[Job], None],
        update_job: Callable[[Job], None],
        restart_job: Callable[[Job], None]
    ) -> 'Pipe':
        p = Pipe(
            model_id=meta_pipe.model_id, 
            pipe_id=meta_pipe.pipe_id, 
            get_job_port=get_job_port,
            app_dir=app_dir,
            complete_job=complete_job,
            update_job=update_job,
            restart_job=restart_job,
            router=router
        )
        local_segments = []
        for model in hosted_models:
            if model.pipe_id == meta_pipe.pipe_id:
                p.segments.append(model)
                local_segments.append(model.process_id)
        p.segments.extend([LlmModel.from_meta(s, app_dir) for s in meta_pipe.segments if s.process_id not in local_segments])
        p.sort_segments()
        return p



def prompt(message: str, default=None, required=False) -> str:
    """Prompt user for input with optional default value."""
    if default is not None:
        display = f"{message} [{default}]: "
    else:
        display = f"{message}: "
    
    while True:
        value = input(display).strip()
        if value == "":
            if default is not None:
                return default
            if required:
                print("  This field is required.")
                continue
            return None
        return value


def prompt_int(message: str, default=None, required=False) -> int:
    """Prompt user for integer input."""
    while True:
        value = prompt(message, default=str(default) if default else None, required=required)
        if value is None:
            return None
        try:
            return int(value)
        except ValueError:
            print("  Please enter a valid number.")


def prompt_float(message: str, default=None, required=False) -> float:
    """Prompt user for float input."""
    while True:
        value = prompt(message, default=str(default) if default else None, required=required)
        if value is None:
            return None
        try:
            return float(value)
        except ValueError:
            print("  Please enter a valid number.")


def prompt_bool(message: str, default=False) -> bool:
    """Prompt user for yes/no input."""
    default_str = "Y/n" if default else "y/N"
    while True:
        value = input(f"{message} [{default_str}]: ").strip().lower()
        if value == "":
            return default
        if value in ("y", "yes", "true", "1"):
            return True
        if value in ("n", "no", "false", "0"):
            return False
        print("  Please enter 'y' or 'n'.")


def prompt_choice(message: str, choices: list, default=None) -> str:
    """Prompt user to select from choices."""
    choices_str = "/".join(choices)
    while True:
        value = prompt(f"{message} ({choices_str})", default=default)
        if value in choices:
            return value
        print(f"  Please choose from: {choices_str}")

def prompt_number_choice(message: str, choices: list, default=None, required=False) -> str:
    print(message)
    for i, choice  in enumerate(choices):
        print(f"[{i}] {choice}")
    if default is not None:
        default_idx = choices.index(default)
    else:
        default_idx = None
    while True:
        try:
            selection = int(prompt("Select number of choice", default=default_idx, required=required))
        except KeyboardInterrupt:
            exit()
        except:
            print("Invalid selection")
            continue
        if selection < 0 or selection >= len(choices):
            print("Invalid selection")
        else:
            return choices[selection]
    return None

def prompt_continue():
    prompt("Press Enter to continue...")



import subprocess

def upgrade_lp():
    print("Upgrading Language Pipes packages")
    subprocess.run(["pip", "install", "language-pipes", "--upgrade"])
    subprocess.run(["pip", "install", "distributed-state-network", "--upgrade"])
    subprocess.run(["pip", "install", "llm-layer-collector", "--upgrade"])


import ssl
import threading
from threading import Thread
from typing import Tuple, Callable
from http.server import HTTPServer, BaseHTTPRequestHandler
from distributed_state_network import DSNode

from language_pipes.util.http import _respond_bytes

class JobHandler(BaseHTTPRequestHandler):
    server: "JobServer"

    def do_POST(self):
        content_length = int(self.headers.get('Content-Length', 0))
        body = self.rfile.read(content_length)
        Thread(target=self.server.cb, args=(body, )).start()
        _respond_bytes(self, b'UP' if not self.server.router.shutting_down else b'DOWN')

    def log_message(self, format, *args):
        pass

class JobServer(HTTPServer):
    router: DSNode
    cb: Callable # Callback into job receiver

    def __init__(
        self, 
        port: int,
        router: DSNode,
        cb: Callable
    ):
        super().__init__(("0.0.0.0", port), JobHandler)
        self.router = router
        self.cb = cb

    def stop(self):
        self.shutdown()
        self.socket.close()

    @staticmethod 
    def start(port: int, router: DSNode, cb: Callable) -> Tuple[Thread, 'JobServer']:
        httpd = JobServer(port, router, cb)
        httpd_thread = threading.Thread(target=httpd.serve_forever, args=())
        httpd_thread.start()

        return httpd_thread, httpd



from enum import Enum
from typing import Optional

class ChatRole(Enum):
    SYSTEM = 'system'
    ASSISTANT = 'assistant'
    USER = 'user'

def str_to_chat_role(role: str) -> Optional[ChatRole]:
    if role == 'system':
        return ChatRole.SYSTEM
    if role == 'assistant':
        return ChatRole.ASSISTANT
    if role == 'user':
        return ChatRole.USER
    return None

class ChatMessage:
    role: ChatRole
    content: str

    def __init__(self, role: ChatRole, content: str):
        self.role = role
        self.content = content

    def to_json(self):
        return {
            'role': self.role.value,
            'content': self.content
        }
    
    @staticmethod
    def from_dict(data) -> Optional["ChatMessage"]:
        role = str_to_chat_role(data['role'])
        if role is None:
            return None
        return ChatMessage(role, data['content'])



import json
from http.server import BaseHTTPRequestHandler

def _respond_bytes(handler: BaseHTTPRequestHandler, data: bytes):
    handler.send_response(200)
    handler.send_header("Content-Type", "application/octet-stream")
    handler.end_headers()
    handler.wfile.write(data)
    handler.wfile.flush()

def _respond_json(handler: BaseHTTPRequestHandler, data):
    response = json.dumps(data).encode("utf-8")
    handler.send_response(200)
    handler.send_header("Content-Type", "application/json")
    handler.end_headers()
    handler.wfile.write(response)
    handler.wfile.flush()

def _send_sse_headers(handler):
    handler.send_response(200)
    handler.send_header("Content-Type", "text/event-stream; charset=utf-8")
    handler.send_header("Cache-Control", "no-cache")
    handler.send_header("Connection", "keep-alive")
    handler.end_headers()

def _send_code(code: int, handler: BaseHTTPRequestHandler, message: str):
    handler.send_response(code)
    handler.send_header("Content-Type", "application/json")
    handler.end_headers()
    handler.wfile.write(message.encode())
    handler.wfile.flush()




import json
from http.server import BaseHTTPRequestHandler, ThreadingHTTPServer
from typing import Callable

from language_pipes.util.oai import oai_chat_complete
from language_pipes.util.http import _send_code

class T:
    complete: Callable

class OAIHttpHandler(BaseHTTPRequestHandler):
    server: T

    def do_POST(self):
        content_length = int(self.headers.get('Content-Length', 0))
        body = self.rfile.read(content_length)
        try:
            data = json.loads(body.decode("utf-8"))
        except json.JSONDecodeError:
            _send_code(400, self, "Invalid JSON")
            return
        
        if 'model' not in data:
            _send_code(400, self, "model parameter is required")
            return

        if 'messages' not in data:
            _send_code(400, self, "messages object parameter is required")
            return

        if len(data['messages']) == 0:
            _send_code(400, self, "messages object must not be empty")
            return
        
        if self.path == '/v1/chat/completions':
            oai_chat_complete(self, self.server.complete, data)

class OAIHttpServer(ThreadingHTTPServer):
    complete: Callable
    
    def __init__(self, port: int, complete: Callable):
        super().__init__(("0.0.0.0", port), OAIHttpHandler)
        self.complete = complete



from typing import List

class HostedModel:
    id: str
    device: str
    max_memory: float # in giga bytes
    load_ends: bool # Loads head and embed of model

    def __init__(self, id: str, device: str, max_memory: float, load_ends: bool):
        self.id = id
        self.device = device
        self.max_memory = max_memory
        self.load_ends = load_ends

    @staticmethod
    def from_dict(data):
        return HostedModel(data['id'], data['device'], data['max_memory'], data['load_ends'])

class ProcessorConfig:
    model_validation: bool
    ecdsa_verification: bool
    max_pipes: int
    job_port: int
    hosted_models: List[HostedModel]

    @staticmethod
    def from_dict(data: dict) -> "ProcessorConfig":
        config = ProcessorConfig()
        config.model_validation = data['model_validation']
        config.ecdsa_verification = data['ecdsa_verification']
        config.max_pipes = data['max_pipes']
        config.job_port = data['job_port']
        config.hosted_models = [HostedModel.from_dict(o) for o in data['hosted_models']]
        return config
    



import os
import torch
from uuid import uuid4
from torch import tensor
from pathlib import Path

from transformers.models.auto.tokenization_auto import AutoTokenizer

from llm_layer_collector import LlmLayerCollector
from llm_layer_collector.auto.auto_rms import AutoRMSNorm
from llm_layer_collector.compute import compute_embedding, compute_head

from language_pipes.util import clone_model
from language_pipes.job_manager.job import ComputeStep, Job
from language_pipes.llm_model.computed import ComputedData
from language_pipes.job_manager.job_data import computationStateToJobData, jobDataToComputationState

class EndModel:
    model_id: str
    process_id: str
    device: str
    input_embedding: torch.nn.Embedding
    norm: AutoRMSNorm
    head: torch.nn.Linear
    collector: LlmLayerCollector

    def __init__(self, app_dir: str, model_id: str, device: str):
        self.model_id = model_id
        self.device = device

        self.process_id = str(uuid4())
        model_dir = str(Path(app_dir) / 'models' / self.model_id)
        if not os.path.exists(model_dir):
            clone_model(model_id, model_dir)
        self.computed = ComputedData(model_dir)
        self.collector = LlmLayerCollector(
            model_dir=os.path.join(model_dir, 'data'),
            cache_file=os.path.join(model_dir, 'cache.json'),
            device=device,
            dtype=torch.float16
        )
        self.tokenizer = AutoTokenizer.from_pretrained(os.path.join(model_dir, 'data'))
    
    def size(self):
        return self.computed.embed_size + self.computed.head_size

    def load(self):
        self.input_embedding = self.collector.load_input_embedding(self.device)
        self.norm = self.collector.load_norm(self.device)
        self.head = self.collector.load_head(self.device)

    def tokenize(self, job: Job):
        prompt = self.tokenizer.apply_chat_template([m.to_json() for m in job.messages], tokenize=False, chat_template=self.tokenizer.chat_template, add_generation_prompt=True)
        input_tokens = [int(t) for t in self.tokenizer.encode(prompt, return_tensors='pt')[0].numpy()]
        job.input_ids = input_tokens
        job.prompt_tokens = len(input_tokens)
        job.next_step()

    def chop_position_embeddings(self, t: torch.Tensor):
        if t is not None:
            return (
                t[0][:, -1:, :],
                t[1][:, -1:, :]
            )

    def compute_embed(self, job: Job):
        if job.current_step != ComputeStep.EMBED:
            self.raise_exception('Invalid step for embedding')
        if self.input_embedding is None:
            self.raise_exception("Input Embedding must be loaded before computation")
        state = None
        if job.data is not None:
            state = jobDataToComputationState(job.data, self.device)
        comp_state = compute_embedding(self.input_embedding, tensor([job.input_ids]).to(self.device), self.collector.config, state)
        if job.current_token > 0:
            comp_state.state = comp_state.state[:, -1:, :]
            if comp_state.causal_mask["full_attention"] is not None:
                comp_state.causal_mask["full_attention"] = comp_state.causal_mask["full_attention"][:, :, -1:, -1:]
            comp_state.position_embeddings = self.chop_position_embeddings(comp_state.position_embeddings)
            comp_state.position_embeddings_local = self.chop_position_embeddings(comp_state.position_embeddings_local)
            comp_state.position_embeddings_global = self.chop_position_embeddings(comp_state.position_embeddings_global)
            
        job.data = computationStateToJobData(comp_state)
        job.next_step()

    def compute_norm(self, job: Job):
        if job.data is None or job.data.state is None:
            self.raise_exception("Cannot compute norm without job data")
        norm = self.norm(job.data.state.to(self.device))
        job.set_norm(norm)
        
    def compute_head(self, job: Job):
        if self.head is None:
            self.raise_exception("Head must be loaded before computation")
        if job.data is None or job.data.state is None:
            self.raise_exception("Cannot compute head without job data")
        head = int(compute_head(self.head, job.data.state.to(self.device))[0][0])
        job.set_output(head, self.collector.config.eos_token_id)
        job.delta = self.tokenizer.decode([job.input_ids[-1]])

    def set_result(self, job: Job):
        res_tokens = job.input_id_tensor()
        job.result = self.tokenizer.decode(res_tokens[job.prompt_tokens:])

    def clean_up(self):
        del self.input_embedding
        del self.norm
        del self.head


import os
import socket
import toml
import argparse
from language_pipes.config import LpConfig
from language_pipes.util.aes import save_new_aes_key
from language_pipes.commands.initialize import interactive_init
from language_pipes.commands.start import start_wizard
from language_pipes.commands.upgrade import upgrade_lp

from language_pipes import LanguagePipes

VERSION = "0.11.2"

def build_parser():
    parser = argparse.ArgumentParser(
        prog="Language Pipes",
        description="Distribute LLMs across multiple systems"
    )

    parser.add_argument("-V", "--version", action="version", version=VERSION)

    subparsers = parser.add_subparsers(dest="command")

    #Upgrade
    subparsers.add_parser("upgrade", help="Upgrade Language Pipes and related packages")

    # Key Generation
    create_key_parser = subparsers.add_parser("keygen", help="Generate AES key")
    create_key_parser.add_argument("output", help="Output file for AES key (default: network.key)", default="network.key")

    # Initialize
    init = subparsers.add_parser("init", help="Create a new configuration file")
    init.add_argument("-o", "--output", default="config.toml")

    # Quick start
    start = subparsers.add_parser("start", help="First-time setup wizard and server start")
    start.add_argument("-c", "--config", default="config.toml", help="Config file path")
    start.add_argument("-k", "--key", default="network.key", help="Network key file path")

    # run command
    run_parser = subparsers.add_parser("serve", help="Start Language Pipes server")
    run_parser.add_argument("-c", "--config", help="Path to TOML config file")
    run_parser.add_argument("-l", "--logging-level", 
        help="Logging verbosity (Default: INFO)",
        choices=["DEBUG", "INFO", "WARNING", "ERROR"]
    )
    run_parser.add_argument("--openai-port", type=int, help="Open AI server port (Default: none)")
    run_parser.add_argument("-d", "--app-data-dir", type=str, help="Application data directory for language pipes (default: ~/.language-pipes)")
    run_parser.add_argument("--node-id", help="Node ID for the network (Required)")
    run_parser.add_argument("--app-dir", type=str, help="Directory to store data for this application")
    run_parser.add_argument("--peer-port", type=int, help="Port for peer-to-peer network (Default: 5000)")
    run_parser.add_argument("--bootstrap-address", help="Bootstrap node address (e.g. 192.168.1.100)")
    run_parser.add_argument("--bootstrap-port", type=int, help="Bootstrap node port for the network (e.g. 8000)")
    run_parser.add_argument("--max-pipes", type=int, help="Maximum amount of pipes to host")
    run_parser.add_argument("--network-key", type=str, help="AES key to access network (Default: network.key)")
    run_parser.add_argument("--model-validation", help="Whether to validate the model weight hashes when connecting to a pipe.", action="store_true")
    run_parser.add_argument("--ecdsa-verification", help="verify legitimacy of sender via ecdsa signed packets" , action="store_true")
    run_parser.add_argument("--job-port", type=int, help="Job receiver port (Default: 5050)")
    run_parser.add_argument("--hosted-models", nargs="*", metavar="MODEL", 
        help="Hosted models as key=value pairs: id=MODEL,device=DEVICE,memory=GB,load_ends=BOOL (e.g., id=Qwen/Qwen3-1.7B,device=cpu,memory=4,load_ends=false)")

    return parser

def apply_overrides(data, args):
    # Environment variable mapping
    env_map = {
        "logging_level": os.getenv("LP_LOGGING_LEVEL"),
        "oai_port": os.getenv("LP_OAI_PORT"),
        "app_dir": os.getenv("LP_APP_DIR"),
        "node_id": os.getenv("LP_NODE_ID"),
        "peer_port": os.getenv("LP_PEER_PORT"),
        "bootstrap_address": os.getenv("LP_BOOTSTRAP_ADDRESS"),
        "bootstrap_port": os.getenv("LP_BOOTSTRAP_PORT"),
        "network_key": os.getenv("LP_NETWORK_KEY"),
        "ecdsa_verification": os.getenv("LP_ECDSA_VERIFICATION"),
        "model_validation": os.getenv("LP_MODEL_VALIDATION"),
        "job_port": os.getenv("LP_JOB_PORT"),
        "max_pipes": os.getenv("LP_MAX_PIPES"),
        "hosted_models": os.getenv("LP_HOSTED_MODELS"),
    }

    def precedence(key, arg, d):
        if arg is not None:
            return arg
        if key in env_map and env_map[key] is not None:
            return env_map[key]
        if key in data:
            return data[key]
        return d
    
    default_app_dir = os.path.expanduser("~") + "/.language-pipes"

    config = {
        "logging_level": precedence("logging_level", args.logging_level, "INFO"),
        "oai_port": precedence("oai_port", args.openai_port, None),
        "app_dir": precedence("app_dir", args.app_dir, default_app_dir),
        "node_id": precedence("node_id", args.node_id, None),
        "peer_port": int(precedence("peer_port", args.peer_port, 5000)),
        "bootstrap_address": precedence("bootstrap_address", args.bootstrap_address, None),
        "bootstrap_port": precedence("bootstrap_port", args.bootstrap_port, 5000),
        "network_key": precedence("network_key", args.network_key, None),
        "ecdsa_verification": precedence("ecdsa_verification", args.ecdsa_verification, False),
        "model_validation": precedence("model_validation", args.model_validation, False),
        "job_port": int(precedence("job_port", args.job_port, 5050)),
        "max_pipes": precedence("max_pipes", args.max_pipes, 1),
        "hosted_models": precedence("hosted_models", args.hosted_models, None),
    }

    if not os.path.exists(default_app_dir):
        Path(default_app_dir).mkdir(parents=True)

    if config["hosted_models"] is None:
        print("Error: hosted_models param must be supplied in config")
        exit()

    if config["node_id"] is None:
        print("Error: node_id param is not supplied in config")
        exit()
    
    if config["oai_port"] is not None:
        config["oai_port"] = int(config["oai_port"])
    
    if config["bootstrap_port"] is not None:
        config["bootstrap_port"] = int(config["bootstrap_port"])

    hosted_models = []
    for m in config["hosted_models"]:
        if type(m) is str:
            # Parse Docker-style key=value pairs: id=X,device=Y,memory=Z,load_ends=W
            model_config = {}
            for pair in m.split(","):
                if "=" not in pair:
                    raise ValueError(f"Invalid format '{pair}' in '{m}'. Expected key=value pairs (e.g., id=Qwen/Qwen3-1.7B,device=cpu,memory=4,load_ends=false)")
                key, value = pair.split("=", 1)
                model_config[key.strip()] = value.strip()
            
            # Validate required keys
            required_keys = {"id", "device", "memory"}
            missing = required_keys - set(model_config.keys())
            if missing:
                raise ValueError(f"Missing required keys {missing} in '{m}'")
            
            hosted_models.append({
                "id": model_config["id"],
                "device": model_config["device"],
                "max_memory": float(model_config["memory"]),
                "load_ends": model_config.get("load_ends", "false").lower() == "true"
            })
        else:
            hosted_models.append(m)

    config["hosted_models"] = hosted_models

    return config

def main(argv = None):
    parser = build_parser()
    args = []
    if argv is None:
        args = parser.parse_args()
    else:
        args = parser.parse_args(argv)

    # Default to "start" command if no command given
    if args.command is None:
        args.command = "start"
        args.config = "config.toml"
        args.key = "network.key"

    if args.command == "keygen":
        key = save_new_aes_key(args.output)
        print(f"✓ Network key generated: {key}")
        print(f"✓ Network key saved to '{args.output}'")
    elif args.command == "upgrade":
        upgrade_lp()
    elif args.command == "init":
        interactive_init(args.output)
    elif args.command == "start":
        try:
            return start_wizard(apply_overrides, VERSION)
        except KeyboardInterrupt:
            exit()
    elif args.command == "serve":
        data = { }
        if args.config is not None:
            with open(args.config, "r", encoding="utf-8") as f:
                data = toml.load(f)
        data = apply_overrides(data, args)
        config = LpConfig.from_dict({
            "logging_level": data["logging_level"],
            "oai_port": data["oai_port"],
            "app_dir": data["app_dir"],
            "router": {
                "node_id": data["node_id"],
                "port": data["peer_port"],
                "aes_key": data["network_key"],
                "bootstrap_nodes": [
                    {
                        "address": data["bootstrap_address"],
                        "port": data["bootstrap_port"]
                    }
                ] if data["bootstrap_address"] is not None else []
            },
            "processor": {
                "max_pipes": data["max_pipes"],
                "model_validation": data["model_validation"],
                "ecdsa_verification": data["ecdsa_verification"],
                "job_port": data["job_port"],
                "hosted_models": data["hosted_models"]
            }
        })

        return LanguagePipes(VERSION, config)
    else:
        parser.print_usage()
        exit(1)

if __name__ == "__main__":
    main()



import random
import requests
from time import time, sleep
from typing import List, Optional, Tuple, Callable

from promise import Promise

from uuid import uuid4
from threading import Thread
from distributed_state_network import DSNode

from language_pipes.util.meta import MetaPipe
from language_pipes.llm_model.end_model import EndModel
from language_pipes.job_manager.router_pipes import RouterPipes
from language_pipes.job_manager.job import Job
from language_pipes.job_manager.pipe import Pipe
from language_pipes.job_manager.enums import JobStatus
from language_pipes.util.chat import ChatMessage
from language_pipes.config.processor import ProcessorConfig
from language_pipes.llm_model import LlmModel
from language_pipes.llm_model.computed import validate_model
from language_pipes.job_manager.pending_job import PendingJob

CHECK_JOB_INTERVAL = 10
EXPIRED_JOB_TIME = 30

class JobManager:
    completed_jobs: List[str]
    jobs_pending: List[PendingJob]
    models: List[LlmModel]
    end_models: List[EndModel]
    app_dir: str
    
    router: DSNode
    
    router_pipes: RouterPipes
    config: ProcessorConfig
    started: bool

    def __init__(self, app_dir: str, router: DSNode, config: ProcessorConfig):
        self.started = False
        self.router = router
        self.config = config
        self.app_dir = app_dir
        self.logger = self.router.logger

        self.jobs_pending = []
        self.completed_jobs = []
        self.models = []
        self.end_models = []
        self.pipes_hosted = []
        self.router_pipes = RouterPipes(router)
        self.router.update_data("job_port", str(self.config.job_port))
        for m in self.config.hosted_models:
            self.host_model(m.id, m.max_memory, m.device, m.load_ends)
        
        self.print_pipes()

        self.started = True
        Thread(target=self.check_stale_jobs, args=( )).start()

    def check_stale_jobs(self):
        while True:
            remove_jobs = []
            for j in self.jobs_pending:
                if time() - j.last_update > EXPIRED_JOB_TIME:
                    remove_jobs.append(j.job.job_id)

            if len(remove_jobs) == 0:
                sleep(CHECK_JOB_INTERVAL)
                continue
        
            for job_id in remove_jobs:
                self.jobs_pending = [j for j in self.jobs_pending if j.job.job_id != job_id]
            
            sleep(CHECK_JOB_INTERVAL)

    def print_pipes(self):
        for p in self.router_pipes.network_pipes():
            p.print(self.logger)


    def raise_exception(self, msg: str):
        self.logger.exception(msg)
        raise Exception(msg)

    def stop(self):
        for m in self.models:
            m.cleanup_tensors()
        for m in self.end_models:
            m.clean_up()
        self.models = []
        self.end_models = []
    
    def get_pipe(self, pipe_id: str) -> Optional[Pipe]:
        meta_pipe = self.router_pipes.network_pipe(pipe_id)
        if meta_pipe is None:
            return None
        return Pipe.from_meta(
            meta_pipe=meta_pipe,
            hosted_models=self.models,
            router=self.router,
            app_dir=self.app_dir,
            get_job_port=self.get_job_port,
            complete_job=self.complete_job,
            update_job=self.update_job,
            restart_job=self.restart_job
        )
    
    def get_job_port(self, router_id: str) -> Optional[int]:
        try:
            return int(self.router.read_data(router_id, 'job_port'))
        except Exception as e:
            self.logger.exception("Error getting job port: %s", e)
            return None

    def update_job(self, job: Job):
        job_id = job.job_id
        if job_id in self.completed_jobs:
            return
        pending_job = self.get_pending_job(job_id)
        if pending_job is None:
            return
        self.logger.info(f'Received job update for {job_id}\n')
        pending_job.last_update = time()
        return pending_job.update(job)

    def complete_job(self, job: Job):
        job_id = job.job_id
        if job_id in self.completed_jobs:
            return
        self.completed_jobs.append(job_id)
        pending_job = self.get_pending_job(job_id)
        if pending_job is None:
            return
        self.logger.info(f'Received job complete for {job_id}\n')
        pending_job.resolve(job)
        self.jobs_pending = [j for j in self.jobs_pending if j.job.job_id != job_id]
      
    def get_model_for_pipe(self, model_id: str, pipe: MetaPipe, device: str, available_memory: int) -> Tuple[int, Optional[LlmModel]]:
        start_memory = available_memory

        new_model: Optional[LlmModel] = LlmModel.from_id(self.app_dir, model_id, self.router.config.node_id, pipe.pipe_id, device)
        computed = new_model.computed
        if self.config.model_validation and len(pipe.segments) > 0 and not validate_model(new_model.computed.to_meta(), pipe.get_computed()):
            self.logger.warning(f'Computed data for model {model_id} does not match')
            return available_memory, None
        
        num_layers_to_load = int(available_memory // computed.avg_layer_size) - 1
        total_layers = new_model.collector.config.num_hidden_layers
        start_layer = pipe.next_start_layer()
        if num_layers_to_load == -1:
            start_layer = -1
            end_layer = -1
        else:
            end_layer = min([start_layer + num_layers_to_load, pipe.next_end_layer(total_layers), new_model.num_hidden_layers]) if start_layer != -1 else -1
            available_memory = available_memory - (end_layer - start_layer + 1) * computed.avg_layer_size

        if num_layers_to_load > -1 and end_layer != -1 and start_layer != -1:
            self.logger.info(f'Using {(start_memory - available_memory) / 10**9:.2f} GB of memory to load model {model_id}')
            new_model.start_layer = start_layer
            new_model.end_layer = end_layer
            new_model.input_embedding = None
            new_model.head = None
            new_model.print()
        else:
            new_model = None
        return available_memory, new_model

    def get_pending_job(self, job_id: str) -> Optional[Job]:
        for j in self.jobs_pending:
            if j.job.job_id == job_id:
                return j
        return None

    def load_end_model(self, model_id: str, device: int):
        model = EndModel(self.app_dir, model_id, device)
        self.end_models.append(model)
        return model

    def host_model(self, model_id: str, max_memory: float, device: str, load_ends: bool):
        available_memory = max_memory * 10 ** 9
        models_to_load: List[LlmModel] = []
        end_model = None
        if load_ends:
            end_model = self.load_end_model(model_id, device)
        
        for pipe_id in [p.pipe_id for p in self.router_pipes.pipes_for_model(model_id, False)]:
            if pipe_id not in self.pipes_hosted and len(self.pipes_hosted) >= self.config.max_pipes:
                break
            loaded = True
            while loaded:
                pipe = self.router_pipes.network_pipe(pipe_id)
                if pipe is None: 
                    break
                available_memory, model = self.get_model_for_pipe(model_id, pipe, device, available_memory)
                loaded = model is not None
                if model is not None:
                    self.pipes_hosted.append(model.pipe_id)
                    self.router_pipes.add_model_to_network(model.to_meta())
                    models_to_load.append(model)

        if len(self.pipes_hosted) < self.config.max_pipes:
            new_pipe = MetaPipe(str(uuid4()), model_id, [])
            self.pipes_hosted.append(new_pipe.pipe_id)
            _, model = self.get_model_for_pipe(model_id, new_pipe, device, available_memory)
            if model is not None:
                self.router_pipes.add_model_to_network(model.to_meta())
                models_to_load.append(model)

        if load_ends:
            end_model.load()

        for m in models_to_load:
            m.load()
            self.router_pipes.update_model(m.to_meta())
            self.models.append(m)

    def get_job_pipe(self, model_id: str) -> Optional[MetaPipe]:
        available_pipes: List[MetaPipe] = []
        for p in self.router_pipes.pipes_for_model(model_id, True):
            if not p.is_loading():
                available_pipes.append(p)
        if len(available_pipes) == 0:
            return None

        return random.choice(available_pipes)
    
    def get_end_model(self, model_id: str) -> Optional[EndModel]:
        for m in self.end_models:
            if m.model_id == model_id:
                return m
        return None

    def restart_job(self, job: Job):
        pipe = self.get_job_pipe(job.model_id)
        if pipe is None:
            job.status = JobStatus.ERROR
            ip = self.router.connection_from_node(job.router_id).address
            port = self.get_job_port(job.router_id)
            cert = self.router.cert_manager.public_path(job.router_id)
            requests.post(f"https://{ip}:{port}", data=job.to_bytes(), headers={ 'Content-Type': 'application/octet-stream' }, verify = cert)
            return
        self.start_job(job.router_id, job.model_id, pipe.pipe_id, job.messages, job.tokens, job.job_id)

    def start_job(
        self, 
        model_id: str, 
        pipe_id: str, 
        messages: List[ChatMessage], 
        tokens: int, 
        job_id: Optional[str] = None
    ) -> Job:
        pipe = self.get_pipe(pipe_id)
        if pipe is None:
            self.raise_exception(f"Could not find pipe {pipe_id}")

        job = Job(self.router.config.node_id, self.router.config.node_id, tokens, messages, pipe_id, model_id)

        if job_id is not None:
            job.job_id = job_id
        
        end_model = self.get_end_model(model_id)
        if end_model is None:
            self.raise_exception(f"Could not find local end model for {model_id}")
            return
        
        end_model.tokenize(job)
        end_model.compute_embed(job)
        first_layer_model = pipe.model_for_job(job)
        if first_layer_model is None:
            self.raise_exception("Could not find appropriate model for processing")
            return
        
        layer_job = job.to_layer_job()
        pipe.send_job(layer_job, first_layer_model.router_id)

        return job

    def complete(
        self, 
        model_id: str, 
        messages: List[ChatMessage], 
        tokens: int, 
        start: Callable,
        update: Callable,
        resolve: Promise,
    ) -> Optional[str]:
        if self.get_end_model(model_id) is None:
            resolve('NO_ENDS')
            return None
        
        network_pipe = self.get_job_pipe(model_id)
        if network_pipe is None:
            resolve('NO_PIPE')
            return None

        job = self.start_job(model_id, network_pipe.pipe_id, messages, tokens)
        start(job)
        self.jobs_pending.append(PendingJob(job, time(), resolve, update))
        return job.job_id



import os
import sys
import time
import pathlib
import requests
import unittest
from typing import List

import openai

cd = pathlib.Path().resolve()

sys.path.append(os.path.join(cd, 'src'))

from language_pipes.cli import main
from language_pipes.util.chat import ChatMessage, ChatRole

MODEL = "Qwen/Qwen3-1.7B"
# MODEL = "Qwen/Qwen3-30B-A3B-Thinking-2507"
# MODEL = "meta-llama/Llama-3.2-1B-Instruct"

def start_node(node_id: str, max_memory: float, peer_port: int, job_port: int, oai_port: int = None, bootstrap_port: int = None):
    args = ["serve", 
        "--node-id", node_id, 
        "--hosted-models", f"id={MODEL},device=cpu,memory={max_memory},load_ends=true", 
        "--peer-port", str(peer_port),
        "--job-port", str(job_port),
        "--app-dir", "./",
        "--model-validation",
    ]
    if oai_port is not None:
        args.extend(["--openai-port", str(oai_port)])
    
    if bootstrap_port is not None:
        args.extend(["--bootstrap-address", "localhost", "--bootstrap-port", str(bootstrap_port)])

    return main(args)

def oai_complete(port: int, messages: List[ChatMessage], retries: int = 0):
    try:
        client = openai.OpenAI(
            api_key="",
            base_url=f"http://127.0.0.1:{port}/v1",
        )
        response = client.chat.completions.create(
            model=MODEL,
            max_completion_tokens=10,
            messages=[m.to_json() for m in messages]
        )
        return response
    except Exception as e:
        print(e)
        if retries < 5:
            time.sleep(5)
            return oai_complete(port, messages, retries + 1)


def oai_stream(port: int, messages: List[ChatMessage], retries: int = 0):
    try:
        client = openai.OpenAI(
            api_key="",
            base_url=f"http://127.0.0.1:{port}/v1",
        )
        stream = client.chat.completions.create(
            model=MODEL,
            # model="gpt-5",
            stream=True,
            max_completion_tokens=10,
            messages=[m.to_json() for m in messages]
        )
        for chunk in stream:
            print(chunk.choices[0].delta.content)
        
    except Exception as e:
        print(e)
        if retries < 5:
            time.sleep(5)
            return oai_complete(port, messages, retries + 1)

class OpenAITests(unittest.TestCase):
    def test_single_node(self):
        start_node("node-1", 5, 5000, 5050, 8000)
        res = oai_complete(8000, [
            ChatMessage(ChatRole.SYSTEM, "You are a helpful assistant"),
            ChatMessage(ChatRole.USER, "Hello, how are you?")
        ])
        print("\"" + res.choices[0].message.content + "\"")
        self.assertTrue(len(res.choices) > 0)

    def test_400_codes(self):
        start_node("node-1", 5, 5000, 5050, 8000)
        messages = [
            ChatMessage(ChatRole.SYSTEM, "You are a helpful assistant"),
            ChatMessage(ChatRole.USER, "Hello, how are you?")
        ]
        res = requests.post("http://localhost:8000/v1/chat/completions", json={
            "messages": [m.to_json() for m in messages]
        })

        self.assertEqual(400, res.status_code)

        res = requests.post("http://localhost:8000/v1/chat/completions", json={
            "model": MODEL
        })

        self.assertEqual(400, res.status_code)

        res = requests.post("http://localhost:8000/v1/chat/completions", json={
            "model": MODEL,
            "messages": []
        })

        self.assertEqual(400, res.status_code)

    def test_double_node(self):
        start_node("node-1", 2, 5000, 5050, 8000)
        time.sleep(5)
        start_node("node-2", 3, 5001, 5051, None, 5000)
        time.sleep(5)
        res = oai_complete(8000, [
            ChatMessage(ChatRole.SYSTEM, "You are a helpful assistant"),
            ChatMessage(ChatRole.USER, "Hello, how are you?")
        ])
        print("\"" + res.choices[0].message.content + "\"")
        self.assertTrue(len(res.choices) > 0)

    def test_stream(self):
        start_node("node-1", 2, 5000, 5050, 8000)
        time.sleep(5)
        start_node("node-2", 3, 5001, 5051, None, 5000)
        time.sleep(5)
        oai_stream(8000, [
            ChatMessage(ChatRole.SYSTEM, "You are a helpful assistant"),
            ChatMessage(ChatRole.USER, "Hello, how are you?")
        ])

    def test_triple_node(self):
        start_node("node-1", 1, 5000, 5050, 8000)
        time.sleep(10)
        start_node("node-2", 1, 5001, 5051, None, 5000)
        time.sleep(10)
        start_node("node-3", 3, 5002, 5052, None, 5000)
        time.sleep(10)
        res = oai_complete(8000, [
            ChatMessage(ChatRole.SYSTEM, "You are a helpful assistant"),
            ChatMessage(ChatRole.USER, "Hello, how are you?")
        ])
        print("\"" + res.choices[0].message.content + "\"")
        self.assertTrue(len(res.choices) > 0)


    def test_reconnect(self):
        start_node("node-1", 1, 5000, 5050, 8000)
        time.sleep(10)
        node2 = start_node("node-2", 1, 5001, 5051, None, 5000)
        time.sleep(10)
        start_node("node-3", 3, 5002, 5052, None, 5000)
        time.sleep(10)
        node2.stop()
        time.sleep(10)
        start_node("node-4", 1, 5004, 5054, None, 5000)
        time.sleep(5)

        res = oai_complete(8000, [
            ChatMessage(ChatRole.SYSTEM, "You are a helpful assistant"),
            ChatMessage(ChatRole.USER, "Hello, how are you?")
        ])
        print("\"" + res.choices[0].message.content + "\"")
        self.assertTrue(len(res.choices) > 0)


if __name__ == '__main__':
    unittest.main()


import os
import re
import toml
from pathlib import Path
from unique_names_generator import get_random_name

from language_pipes.config import LpConfig
from language_pipes.util.aes import save_new_aes_key
from language_pipes.commands.initialize import interactive_init
from language_pipes import LanguagePipes
from language_pipes.util.user_prompts import prompt_bool, prompt, prompt_choice, prompt_number_choice
from language_pipes.util import sanitize_file_name

def start_server(apply_overrides, app_dir: str, config_path: str, version: str):
    print("\nStarting server...\n")
    print("=" * 50)
    
    # Load config and start
    with open(config_path, "r", encoding="utf-8") as f:
        data = toml.load(f)
    
    # Create a minimal args-like object for apply_overrides
    class Args:
        logging_level = None
        openai_port = None
        node_id = None
        peer_port = None
        app_dir = None
        bootstrap_address = None
        bootstrap_port = None
        network_key = None
        model_validation = None
        ecdsa_verification = None
        job_port = None
        max_pipes = None
        hosted_models = None
    
    args = Args()
    data = apply_overrides(data, args)
    
    config = LpConfig.from_dict({
        "logging_level": data["logging_level"],
        "oai_port": data["oai_port"],
        "app_dir": app_dir,
        "router": {
            "node_id": data["node_id"],
            "port": data["peer_port"],
            "credential_dir": str(Path(app_dir) / "credentials"),
            "aes_key_file": data["network_key"],
            "bootstrap_nodes": [
                {
                    "address": data["bootstrap_address"],
                    "port": data["bootstrap_port"]
                }
            ] if data["bootstrap_address"] is not None else []
        },
        "processor": {
            "max_pipes": data["max_pipes"],
            "model_validation": data["model_validation"],
            "ecdsa_verification": data["ecdsa_verification"],
            "job_port": data["job_port"],
            "hosted_models": data["hosted_models"]
        }
    })

    try:
        return LanguagePipes(version, config)
    except KeyboardInterrupt:
        return
    except Exception as e:
        print(e)
        print(e.stack)

def new_config(app_dir: str):
    raw_name = prompt("Name of new configuration", required=True)
    file_name = sanitize_file_name(raw_name)
    if not file_name:
        print("Invalid file name")
        return
    
    config_dir = str(Path(app_dir) / "configs")
    file_path = str(Path(app_dir) / "configs" / file_name)

    if not os.path.exists(config_dir):
        Path.mkdir(config_dir)
    interactive_init(file_path)

def delete_config(app_dir: str):
    config_path = select_config(app_dir)
    os.remove(config_path)
    print("Configuration deleted")

def get_config_files(config_dir: str):
    if os.path.exists(config_dir):
        return [f.replace(".toml", "") for f in os.listdir(config_dir)]
    else:
        return []

def select_config(app_dir: str):
    config_dir = str(app_dir / "configs")

    if not os.path.exists(config_dir):
        Path(config_dir).mkdir(parents=True)

    existing_configs = get_config_files(config_dir)

    if len(existing_configs) > 0:
        load_config = prompt_number_choice("Load Configuration", existing_configs, required=True)
        if load_config is None:
            exit()
        load_config = load_config + ".toml"
    else:
        print("No configs found...")
        return

    return str(Path(config_dir) / load_config)

def view_config(app_dir: str):
    config_path = select_config(app_dir)
    with open(config_path, 'r', encoding='utf-8') as f:
        data = toml.load(f)
    print("\n\n" + "="*50 + "\n")
    print(toml.dumps(data))
    print("="*50 + "\n\n")

def start_wizard(apply_overrides, version: str):
    print("""         
==============================================================================

 | |                                              |  __ (_)                
 | |     __ _ _ __   __ _ _   _  __ _  __ _  ___  | |__) | _ __   ___  ___ 
 | |    / _` | '_ \ / _` | | | |/ _` |/ _` |/ _ \ |  ___/ | '_ \ / _ \/ __|
 | |___| (_| | | | | (_| | |_| | (_| | (_| |  __/ | |   | | |_) |  __/\__ \\
 |______\__,_|_| |_|\__, |\__,_|\__,_|\__, |\___| |_|   |_| .__/ \___||___/
                     __/ |             __/ |              | |              
                    |___/             |___/               |_|      

==============================================================================

- Made with <3 by Erin
""")

    default_base_path = Path(os.path.expanduser("~") ) / ".language-pipes"
    app_dir = default_base_path
    if not os.path.exists(default_base_path):
        app_dir = prompt("Where should we store application data?", required=True, default=str(default_base_path))
    
    if not os.path.exists(app_dir):
        Path(app_dir).mkdir(parents=True)
        print(f"Created directory: {app_dir}")

    while True: # TODO Add model list
        main_menu_cmd = prompt_number_choice("Main Menu", [
            "View Config",
            "Load Config",
            "Create Config",
            "Delete Config"
        ])

        config_dir = str(Path(app_dir) / "configs")
        match main_menu_cmd:
            case "Load Config":
                if len(get_config_files(config_dir)) == 0:
                    print("No configs found...")
                    continue
                config_path = select_config(app_dir)
                return start_server(apply_overrides, app_dir, config_path, version)
            case "View Config":
                if len(get_config_files(config_dir)) == 0:
                    print("No configs found...")
                    continue
                view_config(app_dir)
            case "Create Config":
                new_config(app_dir)
            case "Delete Config":
                if len(get_config_files(config_dir)) == 0:
                    print("No configs found...")
                    continue
                delete_config(app_dir)


import os
import socket
import toml
from unique_names_generator import get_random_name

from language_pipes.util.aes import generate_aes_key
from language_pipes.util.user_prompts import prompt, prompt_bool, prompt_choice, prompt_float, prompt_int

def get_default_node_id() -> str:
    return socket.gethostname()

def interactive_init(output_path: str):
    """Interactively create a configuration file."""
    print("\n" + "=" * 50)
    print("  Configuration Setup")
    print("=" * 50)
    print("\nThis wizard will help you create a new configuration.")

    config = {}

    # === Required Settings ===
    print("--- Required Settings ---\n")

    print("The node ID is a unique name that identifies this computer on the")
    print("Language Pipes network. Other nodes will use this to route jobs.\n")
    config["node_id"] = prompt(
        "Node ID",
        default=get_default_node_id(),
        required=True
    )

    # === Model Configuration ===
    print("\n--- Model Configuration ---\n")
    print("Language Pipes hosts parts of large language models distributed across")
    print("multiple machines. You need to specify at least one model to host.")
    print("Models are automatically downloaded from HuggingFace by their ID.\n")

    hosted_models = []
    while True:
        print(f"  Model #{len(hosted_models) + 1}:")
        
        print("    Enter the HuggingFace model ID. This is the identifier used on")
        print("    huggingface.co (e.g., 'Qwen/Qwen3-1.7B', 'meta-llama/Llama-3.2-1B-Instruct').")
        model_id = prompt(
            "    Model ID",
            default="Qwen/Qwen3-1.7B" if len(hosted_models) == 0 else None,
            required=len(hosted_models) == 0
        )
        
        if model_id is None:
            break
        
        print("\n    Select the compute device for this model. Use 'cpu' for CPU-only,")
        print("    or 'cuda:0', 'cuda:1', etc. for specific GPUs.")
        device = prompt(
            "    Device",
            default="cpu",
            required=True
        )
        
        print("\n    Specify the maximum RAM/VRAM (in GB) this node should use for the model.")
        print("    The model layers will be loaded until this limit is reached.")
        max_memory = prompt_float(
            "    Max memory (GB)",
            default=4,
            required=True
        )
        
        print("\n    The 'ends' of a model are the embedding layer (input) and the output head.")
        print("    At least one node in the network needs these loaded to process requests.")
        print("    If this is the only node or the first/last in a chain, enable this.")
        load_ends = prompt_bool(
            "    Load embedding/output layers",
            default=True
        )
        
        hosted_models.append({
            "id": model_id,
            "device": device,
            "max_memory": max_memory,
            "load_ends": load_ends
        })
        
        print()
        if not prompt_bool("Add another model?", default=False):
            break

    config["hosted_models"] = hosted_models

    # === API Server ===
    print("\n--- API Server ---\n")

    print("Language Pipes can expose an OpenAI-compatible HTTP API, allowing you to")
    print("use standard OpenAI client libraries (Python, JavaScript, curl, etc.)")
    print("to interact with your distributed model.\n")
    if prompt_bool("Enable OpenAI-compatible API server?", default=True):
        print("\n  Choose a port for the API server. Clients will connect to")
        print("  http://<this-machine>:<port>/v1/chat/completions")
        config["oai_port"] = prompt_int(
            "  API port",
            default=8000,
            required=True
        )

    # === Network Configuration ===
    print("\n--- Network Configuration ---\n")

    print("Language Pipes uses a peer-to-peer network to coordinate between nodes.")
    print("The first node starts fresh; additional nodes connect to an existing node.\n")
    is_first_node = prompt_bool(
        "Is this the first node in the network?",
        default=True
    )

    if not is_first_node:
        print("\n  Enter the IP address of an existing node on the network.")
        print("  This node will connect to it to join the distributed network.")
        config["bootstrap_address"] = prompt(
            "  Bootstrap node IP",
            required=True
        )
        print("\n  Enter the peer port of the bootstrap node (default is 5000).")
        config["bootstrap_port"] = prompt_int(
            "  Bootstrap node port",
            default=5000,
            required=True
        )

    print("\nThe peer port is used for network coordination and discovery.")
    print("Other nodes will connect to this port to join the network.")
    config["peer_port"] = prompt_int(
        "Peer port",
        default=5000
    )

    print("\nThe job port is used for transferring computation data between nodes")
    print("during model inference (hidden states, embeddings, etc.).")
    config["job_port"] = prompt_int(
        "Job port",
        default=5050
    )
    
    print("\nThe network key is an AES encryption key shared by all nodes.")
    print("It encrypts communication and prevents unauthorized access.")
    print("There will be no encryption between nodes if the default value is selected.")
    print("")
    encrypt_traffic = prompt_bool("Encrypt network traffic", default=False)
    if encrypt_traffic:
        config["network_key"] = prompt(
            "Network key",
            default="Generate new key"
        )

        if config["network_key"] == "Generate new key":
            key = generate_aes_key().hex()
            config["network_key"] = key
            print(f"Generated new key: {key}")
            print("Note: Save this key somewhere and supply it to other nodes on the network")
    else:
        config["network_key"] = None

    # === Advanced Options ===
    print("\n--- Advanced Options ---\n")

    print("Advanced options include logging verbosity, security features, and limits.")
    if prompt_bool("Configure advanced options?", default=False):
        print("\n  Controls how much information is printed to the console.")
        print("  DEBUG shows everything, ERROR shows only critical issues.")
        config["logging_level"] = prompt_choice(
            "  Logging level",
            ["DEBUG", "INFO", "WARNING", "ERROR"],
            default="INFO"
        )
        
        print("\n  Limits how many model 'pipes' (distributed model instances)")
        print("  this node will participate in simultaneously.")
        config["max_pipes"] = prompt_int(
            "  Max pipes",
            default=1
        )
        
        print("\n  When enabled, nodes verify that model weight hashes match")
        print("  to ensure all nodes are running the exact same model.")
        config["model_validation"] = prompt_bool(
            "  Validate model hashes?",
            default=False
        )
        
        print("\n  ECDSA verification signs each job packet cryptographically,")
        print("  ensuring jobs only come from authorized nodes in the pipe.")
        config["ecdsa_verification"] = prompt_bool(
            "  Enable ECDSA signing?",
            default=False
        )

    # === Write Config File ===
    print("\n" + "=" * 50)
    print("  Configuration Summary")
    print("=" * 50 + "\n")

    # Build clean config (remove None values)
    clean_config = {k: v for k, v in config.items() if v is not None and k != "hosted_models"}
    clean_config["hosted_models"] = config["hosted_models"]

    # Generate TOML preview
    preview = toml.dumps(clean_config)
    print(preview)

    # Check if file exists
    if os.path.exists(output_path):
        if not prompt_bool(f"  '{output_path}' already exists. Overwrite?", default=False):
            print("Aborted.")
            return
    
    with open(output_path, 'w', encoding='utf-8') as f:
        toml.dump(clean_config, f)
    
    print(f"\n✓ Configuration saved")


import os
from io import BytesIO
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives.kdf.pbkdf2 import PBKDF2HMAC
from cryptography.hazmat.primitives import hashes, padding as sym_padding
from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes

def generate_aes_key() -> bytes:
    key = PBKDF2HMAC(
        algorithm=hashes.SHA256(),
        length=16,
        salt=os.urandom(16),
        iterations=100000,
        backend=default_backend()
    ).derive(os.urandom(128))
    iv = os.urandom(16)
    bts = BytesIO()
    bts.write(iv)
    bts.write(key)
    return bts.getvalue()

def save_new_aes_key(file_path: str) -> str:
    key = generate_aes_key()
    key_hex = key.hex()
    with open(file_path, 'w', encoding='utf-8') as f:
        f.write(key_hex)
    return key_hex



import io
import re
import os
import shutil
from pathlib import Path
import json
import base64
import ctypes
import subprocess
from uuid import UUID
from hashlib import sha256
from threading import Thread

import torch

def uuid_to_bytes(uid: str) -> bytes:
    return UUID(hex=uid).bytes

def bytes_to_uuid(b: bytes) -> str:
    return str(UUID(bytes=b))

def int_to_bytes(i: int) -> bytes:
    return i.to_bytes(4, 'little', signed=False)

def bytes_to_int(b: bytes) -> int:
    return int.from_bytes(b, 'little', signed=False)

def tensor_to_bytes(t: torch.Tensor) -> bytes:
    bts = io.BytesIO()
    torch.save(t, bts)
    return bts.getvalue()

def bytes_to_tensor(b: bytes) -> torch.Tensor:
    if b == b'':
        return None
    return torch.load(io.BytesIO(b))

def get_tensor_byte_string(t: torch.Tensor) -> str:
    bts = tensor_to_bytes(t)
    return base64.b64encode(bts).decode('utf-8')

def get_hash(data: str) -> str:
    hash = sha256(data.encode())
    return hash.hexdigest()

def get_dict_hash(data: dict):
    return get_hash(json.dumps(data))

def size_of_tensor(t: torch.Tensor):
    return t.element_size() * t.nelement()

def tensor_hash(t: torch.Tensor) -> str:
    return get_hash(get_tensor_byte_string(t))

def stop_thread(thread: Thread):
    thread_id = thread.ident
    res = ctypes.pythonapi.PyThreadState_SetAsyncExc(thread_id,
            ctypes.py_object(SystemExit))
    if res > 1:
        ctypes.pythonapi.PyThreadState_SetAsyncExc(thread_id, 0)
        print('Exception raise failure')

def clone_model(model_id: str, model_dir: str):
    repo_url = f"https://huggingface.co/{model_id}"
    clone_dir = f"{model_dir}/data"

    if not os.path.exists(clone_dir):
        Path(clone_dir).mkdir(parents=True)
    try:
        subprocess.run(["git", "clone", repo_url, clone_dir])
        subprocess.run(["git", "lfs", "install"], cwd=clone_dir, check=True)
        subprocess.run(["git", "lfs", "pull"], cwd=clone_dir, check=True)
    except Exception as e:
        if os.path.exists(model_dir):
            shutil.rmtree(model_dir)
        print(e)
        print("Git LFS Error occurred: please ensure that git-lfs is installed")
        exit()


def sanitize_file_name(raw_name: str):
    return re.sub(r'[<>:"/\\|?*\x00-\x1f]', '_', raw_name).strip().strip('.') + ".toml"


from time import sleep
from threading import Thread
from typing import Callable, Optional, List
from distributed_state_network import DSNode

from language_pipes.job_manager.pipe import Pipe
from language_pipes.job_manager.job import Job
from language_pipes.llm_model.end_model import EndModel
from language_pipes.job_manager.enums import ComputeStep, JobStatus
from language_pipes.handlers.job import JobServer
from language_pipes.util import stop_thread
from language_pipes.job_manager.layer_job import LayerJob
from language_pipes.config.processor import ProcessorConfig
from language_pipes.job_manager.pending_job import PendingJob

class JobReceiver:
    port: int
    public_key_file: str
    private_key_file: str
    ecdsa_verification: bool
    router: DSNode
    pending_jobs: List[LayerJob]
    get_pipe: Callable[[str], Optional[Pipe]]
    get_end_model: Callable[[str], Optional[EndModel]]
    restart_job: Callable[[Job], None]

    def __init__(
            self, 
            config: ProcessorConfig,
            router: DSNode,
            get_pipe: Callable[[str], Optional[Pipe]],
            get_end_model: Callable[[str], Optional[EndModel]],
            get_pending_job: Callable[[str], Optional[PendingJob]],
            restart_job: Callable[[Job], None]
    ):
        self.router = router
        self.get_pipe = get_pipe
        self.get_end_model = get_end_model
        self.restart_job = restart_job
        self.get_pending_job = get_pending_job
        self.pending_jobs = []
        self.ecdsa_verification = config.ecdsa_verification

        thread, httpd = JobServer.start(config.job_port, self.router, self.receive_data)
        self.thread = thread
        self.httpd = httpd
        Thread(target=self.job_runner, args=()).start()
        router.logger.info(f"Started Job Receiver on port {config.job_port}")

    def job_runner(self):
        while True:
            try:
                if self.router.shutting_down:
                    return
                
                if len(self.pending_jobs) == 0:
                    sleep(0.1)
                    continue
                
                layer_job = self.pending_jobs[-1]
                self.pending_jobs.pop()
                
                pipe = self.get_pipe(layer_job.pipe_id)
                end_model = self.get_end_model(pipe.model_id)

                if pipe is None or not pipe.is_complete():
                    self.restart_job(layer_job)
                    continue
                
                if layer_job.done:
                    job = self.get_pending_job(layer_job.job_id).job
                    job.current_step = ComputeStep.NORM
                    job.data = layer_job.data
                    end_model.compute_norm(job)
                    end_model.compute_head(job)
                    if job.status == JobStatus.COMPLETED:
                        end_model.set_result(job)
                        pipe.complete_job(job)
                        continue
                    else:
                        if not pipe.update_job(job):
                            continue
                        end_model.compute_embed(job)
                        layer_job = job.to_layer_job()

                model = pipe.model_for_job(layer_job)
                model.process_job(layer_job)

                if layer_job.done:
                    pipe.send_job(layer_job, layer_job.origin_node_id)
                else:
                    model = pipe.model_for_job(layer_job)
                    pipe.send_job(layer_job, model.router_id)
            except Exception as e:
                print(e)

    def receive_data(self, data: bytes):
        job = LayerJob.from_bytes(data)
        
        for j in self.pending_jobs:
            if j.job_id == job.job_id:
                return

        self.pending_jobs.insert(0, job)

    def stop(self):
        self.httpd.stop()
        stop_thread(self.thread)